from distutils.core import setup
setup(
  name = 'Pynac',
  packages = ['Pynac'], # this must be the same as the name above
  version = '0.9.4',
  description = 'A Python wrapper around the Dynac charged particle simulator',
  author = 'Stephen Molloy',
  author_email = 'sdmolloy@gmail.com',
  url = 'https://github.com/se-esss-litterbox/Pynac', # use the URL to the github repo
  download_url = 'https://github.com/se-esss-litterbox/Pynac/archive/0.9.4.tar.gz',
  keywords = ['simulation', 'particle', 'dynamics'], # arbitrary keywords
  classifiers = [],
)
